// Optimierungsroutinen moldable, beliebiger Modulabhängigkeitsgraph
// Graphzerlegung in Bäume und Behandlung der Bäume als serien-parallele
// Graphen
//
// optimiert Module deren Laufzeiten in der Datei daten stehen
// (i-te Zeile j-te Spalte enthält Laufzeit vom i-ten Modul
//  mit j Prozessoren)
// oder mit Hilfe der übergebenen Funktion laufzeit
// (modulnummer>=1, prozessorenanzahl>=1) ermittelt werden können
//
// Aufruf:
//   optimiere_datei(
//     char* Ergebnisdatei,
//     char* Laufzeitendatei,
//     char* Graphendatei )
// oder:
//   optimiere_formel(
//     double** Startzeit der Module,
//     int**    Modulbreiten,
//     int**    linker Prozessor jedes Moduls,
//     graph*   Modulabhängigkeitsgraph,
//     int      Maximalprozessorenanzahl,
//     double   ((*laufzeit)( modulnummer, prozessorenanzahl )) )
//   Der Speicher für die Startzeiten, Breiten und linker Prozessor
//   wird reserviert
//
// Fehler: 1->zu wenig Speicher
//         2->Eingabedatei mit Laufzeiten nicht gefunden / fehlerhaft
//         3->Eingabedatei mit Graphen nicht gefunden / fehlerhaft /
//            Graph enthält Zyklus
//         4->kann Ergebnisdatei nicht anlegen / Schreibfehler

#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include "../prgx/graphen.h"
#include "../prgx/listen.h"
#include "../prgx/load.h"
#include "../prgx/save.h"
#include "../prgx/suchsort.h"


// berechnet die parallele Vereinigung zweier Module mit den
// Laufzeiten za und zb
// in typ[n*z+i] mit i=0..n-1 wird der Vereinigungstyp mit i+1
// Prozessoren festgehalten
// Ergebnis liegt in za
void verein_parallel( int z, int n, int* typ, double* za, double* zb )
{
  int i, j, x, y;
  z--;
  // Dimension n wegen Bestimmung vom Wechsel extra betrachten
  za[n-1] += zb[n-1];
  typ[ z * n + n - 1 ] = 0;
  x = 0; // Stelle mit Wechsel der Größen von za und zb suchen
  y = n;
  while( x + 1 < y )
  {
    j = ( x + y ) / 2;
    if( za[j-1] <= zb[n-j-1] ) // hier ist j immer >=1 und <=n-1
    {
      y = j;
    }
    else
    {
      x = j;
    }
  }
  // wenn 1 <= x <= n-2 gilt za[x] <= zb[n-x-2] und za[x-1] > zb[n-x-1]
  // wenn x == 0 gilt für alle i: za[i] <= za[0] <= zb[n-2]
  // wenn x == n-1 gilt für alle i: za[n-2] > zb[0] >= zb[i]
  if( x == 0 ) // za[x-1] existiert nicht
  {
    if( za[n-1] >= zb[n-2] ) // parallele Ausführung besser ?
    {
      za[n-1] = zb[n-2];
      typ[ z * n + n - 1 ] = 1;
    }
  }
  else // x >= 1
  {
    if( x == n-1 )
    {
      if( za[n-1] >= za[n-2] )
      {
        za[n-1] = za[n-2];
        typ[ z * n + n - 1 ] = n-1;
      }
      x--; // nächster Wert
    }
    else // 1 <= x <= n-2
    {
      // Minimum aus beiden Möglichkeiten suchen
      if( za[x-1] <= zb[n-x-2] )
      {
        if( za[n-1] >= za[x-1] )
        {
          za[n-1] = za[x-1];
          typ[ z * n + n - 1 ] = x;
        }
        x--; // nächster Wert
      }
      else
      {
        if( za[n-1] >= zb[n-x-2] )
        {
          za[n-1] = zb[n-x-2];
          typ[ z * n + n - 1 ] = x + 1;
        }
      }
    }
  }
  for( i=n-2; i>=1; i-- )
  {
    za[i] += zb[i];
    typ[ z * n + i ] = 0;
    if( x == 0 ) // za[x-1] existiert nicht
    {
      if( za[i] >= zb[i-1] ) // parallele Ausführung besser ?
      {
        za[i] = zb[i-1];
        typ[ z * n + i ] = 1;
      }
    }
    else // x >= 1
    {
      if( x == i )
      {
        if( za[i] >= za[i-1] )
        {
          za[i] = za[i-1];
          typ[ z * n + i ] = i;
        }
        x--; // nächster Wert
      }
      else // 1 <= x <= i-1
      {
        // Minimum aus beiden Möglichkeiten suchen
        if( za[x-1] <= zb[i-x-1] )
        {
          if( za[i] >= za[x-1] )
          {
            za[i] = za[x-1];
            typ[ z * n + i ] = x;
          }
          x--; // nächster Wert
        }
        else
        {
          if( za[i] >= zb[i-x-1] )
          {
            za[i] = zb[i-x-1];
            typ[ z * n + i ] = x + 1;
          }
        }
      }
    }
  }
  za[0] += zb[0]; // bei einem Prozessor gibt es nur eine Möglichkeit
  typ[ z * n ] = 0;
}


// ---------------------- Formelorientierte Routinen -------------------
#ifdef FORMEL

int verein_modul_formel(
  graph*  g,
  int     w,
  int*    indeg,
  int*    typ,
  int     n,
  double* za,
  double  ((*laufzeit)( int modulnummer, int prozessorenanzahl )) );

// berechnet die Zeiten za vom fiktiven Modul w (rekursiver Abstieg
// zum Ausbalancieren des Baumes)
// g ist der Abhängigkeitsgraph und indeg seine fiktiven Eingangsgrade;
// in typ wird die durchgeführte Vereinigung gespeichert: es sei z
// die Nummer des i-ten Kindes, so stehen ab typ[(z-1)*n] die
// Vereinigungsdaten des i-ten inneren Knoten bzgl. DFS des
// Konstruktionsbaumes der Kinder von w
// Fehler: 1->zu wenig Speicher
int verein_modul_formel_bal1(
  graph*        g,
  int           w,
  int*          indeg,
  int*          typ,
  int           n,
  double*       za,
  double        ((*laufzeit)( int modulnummer, int prozessorenanzahl )),
  int           anz,
  knotenliste** typ_nr )
{
  double* zb;
  int z;
  if( anz == 1 )
  {
    do
    {
      z = next_kind( g, w );
    } while( indeg[z-1] != 1 );
    return verein_modul_formel( g, z, indeg, typ, n, za, laufzeit );
  }
  zb = (double*)malloc( n * sizeof( double ) );
  if( zb == NULL )
  {
    return 1;
  }
  // suchen des Moduls für die Vereinigungsinformationen typ
  *typ_nr = swap_kind( g, w, *typ_nr );
  do
  {
    z = next_kind( g, w );
  } while( indeg[z-1] != 1 );
  *typ_nr = swap_kind( g, w, *typ_nr );
  if( verein_modul_formel_bal1( g, w, indeg, typ, n, za, laufzeit,
                                anz - anz/2, typ_nr ) != 0 )
  {
    return 1;
  }
  if( verein_modul_formel_bal1( g, w, indeg, typ, n, zb, laufzeit,
                                anz/2, typ_nr ) != 0 )
  {
    return 1;
  }
  verein_parallel( z, n, typ, za, zb );
  free( zb );
  return 0;
}

// berechnet die Zeiten za von den Modulen in der Liste und vereinigt
// sie parallel (rekursiver Abstieg zum Ausbalancieren des Baumes)
// g ist der Abhängigkeitsgraph und indeg seine fiktiven Eingangsgrade;
// in typ wird die durchgeführte Vereinigung gespeichert
// Fehler: 1->zu wenig Speicher
int verein_modul_formel_bal2(
  graph*    g,
  intliste* liste,
  int*      indeg,
  int*      typ,
  int       n,
  double*   za,
  double    ((*laufzeit)( int modulnummer, int prozessorenanzahl )),
  int       anz,
  il_pos*   typ_nr )
{
  double* zb;
  int z;
  if( anz == 1 )
  {
    z = next_il( liste );
    return verein_modul_formel( g, z, indeg, typ, n, za, laufzeit );
  }
  zb = (double*)malloc( n * sizeof( double ) );
  if( zb == NULL )
  {
    return 1;
  }
  // suchen des Moduls für die Vereinigungsinformationen typ
  swap_pos_il( liste, typ_nr );
  z = next_il( liste );
  swap_pos_il( liste, typ_nr );
  if( verein_modul_formel_bal2( g, liste, indeg, typ, n, za, laufzeit,
                                anz - anz/2, typ_nr ) != 0 )
  {
    return 1;
  }
  if( verein_modul_formel_bal2( g, liste, indeg, typ, n, zb, laufzeit,
                                anz/2, typ_nr ) != 0 )
  {
    return 1;
  }
  verein_parallel( z, n, typ, za, zb );
  free( zb );
  return 0;
}

// berechnet die Zeiten za vom fiktiven Modul w;
// g ist der Abhängigkeitsgraph und indeg seine fiktiven Eingangsgrade;
// in typ wird die durchgeführte Vereinigung gespeichert
// Fehler: 1->zu wenig Speicher
int verein_modul_formel(
  graph*  g,
  int     w,
  int*    indeg,
  int*    typ,
  int     n,
  double* za,
  double  ((*laufzeit)( int modulnummer, int prozessorenanzahl )) )
{
  int i, z;
  int anz;
  knotenliste* typ_nr;
  reset_kind( g, w );
  anz = 0; // mögliche Kinder zählen
  while( 1 )
  {
    z = next_kind( g, w );
    if( z == 0 ) // keine Kinder mehr
    {
      break;
    }
    if( indeg[z-1] == 1 ) // verwendbares Kind
    {
      anz++;
    }
  }
  if( anz == 0 )
  { // Blatt in diesem Teilbaum erreicht; Modullaufzeiten zurückgeben
    for( i=0; i<n; i++ )
    {
      za[i] = laufzeit( w, i+1 );
    }
    return 0;
  }
  reset_kind( g, w );
  typ_nr = swap_kind_get( g, w );
  if( verein_modul_formel_bal1( g, w, indeg, typ, n, za, laufzeit,
                                anz, &typ_nr ) != 0 )
  {
    return 1;
  }
  // sequentielle Vereinigung mit Modul w
  for( i=0; i<n; i++ )
  {
    za[i] += laufzeit( w, i+1 );
  }
  return 0;
}


// Vorwärtsdeklaration
double setze_zeiten_formel(
  graph*    g,
  int       w,
  int*      indeg,
  int*      typ,
  intliste* liste,
  int       n,
  double    ((*laufzeit)( int modulnummer, int prozessorenanzahl )),
  double*   startzeit,
  int*      breite,
  int*      linker,
  int       li,
  int       re,
  double    st );

// arbeitet Elemente der nachfolgerliste baumartig ab und berechnet
// die Startzeiten
// Fehler: <0 zu wenig Speicher
double setze_zeiten_formel_bal1(
  graph*    g,
  intliste* nachfolger,
  int*      indeg,
  int*      typ,
  intliste* liste,
  int       n,
  double    ((*laufzeit)( int modulnummer, int prozessorenanzahl )),
  double*   startzeit,
  int*      breite,
  int*      linker,
  int       li,
  int       re,
  double    st,
  int       anz )
{
  int z, h;
  double sr;
  if( anz == 1 ) // nur noch ein Kind
  {
    z = del_anf_il( nachfolger );
    return setze_zeiten_formel( g, z, indeg, typ, liste, n, laufzeit,
                                startzeit, breite, linker, li, re, st );
  }
  z = next_il( nachfolger ); // Modul mit Vereinigungsinformationen
  if( typ[ (z-1)*n + re - li - 1 ] == 0 ) // seq. Vereinigung
  {
    st = setze_zeiten_formel_bal1( g, nachfolger, indeg, typ, liste, n,
                                   laufzeit, startzeit, breite, linker,
                                   li, re, st, anz - anz/2 );
    if( st < 0 )
    {
      return -1;
    }
    return setze_zeiten_formel_bal1( g, nachfolger, indeg, typ, liste,
                                     n, laufzeit, startzeit, breite,
                                     linker, li, re, st, anz/2 );
  }
  // parallele Vereinigung
  h = li + typ[ (z-1)*n + re - li - 1 ];
  sr = setze_zeiten_formel_bal1( g, nachfolger, indeg, typ, liste, n,
                                 laufzeit, startzeit, breite, linker,
                                 li, h, st, anz - anz/2 );
  if( sr < 0 )
  {
    return -1;
  }
  st = setze_zeiten_formel_bal1( g, nachfolger, indeg, typ, liste, n,
                                 laufzeit, startzeit, breite, linker,
                                 h, re, st, anz/2 );
  if( st < 0 )
  {
    return -1;
  }
  if( sr > st )
  {
    return sr;
  }
  return st;
}

// arbeitet die Kinder von baumartig ab und berechnet die Startzeiten
// Fehler: -1->zu wenig Speicher
double setze_zeiten_formel_bal2(
  graph*        g,
  int           w,
  int*          indeg,
  int*          typ,
  intliste*     liste,
  int           n,
  double        ((*laufzeit)( int modulnummer, int prozessorenanzahl )),
  double*       startzeit,
  int*          breite,
  int*          linker,
  int           li,
  int           re,
  double        st,
  int           anz,
  knotenliste** typ_nr )
{
  int z, h;
  double sr;
  if( anz == 1 ) // nur noch ein Kind
  {
    do
    {
      z = next_kind( g, w );
    } while( indeg[z-1] != 1 );
    return setze_zeiten_formel( g, z, indeg, typ, liste, n, laufzeit,
                                startzeit, breite, linker, li, re, st );
  }
  // Modul mit Vereinigungsinformationen suchen
  *typ_nr = swap_kind( g, w, *typ_nr );
  do
  {
    z = next_kind( g, w );
  } while( indeg[z-1] != 1 );
  *typ_nr = swap_kind( g, w, *typ_nr );
  if( typ[ (z-1)*n + re - li - 1 ] == 0 ) // seq. Vereinigung
  {
    st = setze_zeiten_formel_bal2( g, w, indeg, typ, liste, n, laufzeit,
                                   startzeit, breite, linker, li, re,
                                   st, anz - anz/2, typ_nr );
    if( st < 0 )
    {
      return -1;
    }
    return setze_zeiten_formel_bal2( g, w, indeg, typ, liste, n,
                                     laufzeit, startzeit, breite,
                                     linker, li, re, st, anz/2,
                                     typ_nr );
  }
  // parallele Vereinigung
  h = li + typ[ (z-1)*n + re - li - 1 ];
  sr = setze_zeiten_formel_bal2( g, w, indeg, typ, liste, n, laufzeit,
                                 startzeit, breite, linker, li, h, st,
                                 anz - anz/2, typ_nr );
  if( sr < 0 )
  {
    return -1;
  }
  st = setze_zeiten_formel_bal2( g, w, indeg, typ, liste, n, laufzeit,
                                 startzeit, breite, linker, h, re, st,
                                 anz/2, typ_nr );
  if( st < 0 )
  {
    return -1;
  }
  if( sr > st )
  {
    return sr;
  }
  return st;
}

// speichert das Ergebnis indem der rekursive Abstieg, wie in
// verein_modul berechnet, durchgeführt wird
// Rückgabe: >0 Endzeit des fiktiven Knotens w bei Startzeit st und
//              Breite re-li und linken Prozessor li
//           <0 zu wenig Speicher
double setze_zeiten_formel(
  graph*    g,
  int       w,
  int*      indeg,
  int*      typ,
  intliste* liste,
  int       n,
  double    ((*laufzeit)( int modulnummer, int prozessorenanzahl )),
  double*   startzeit,
  int*      breite,
  int*      linker,
  int       li,
  int       re,
  double    st )
{
  int anz, z;
  knotenliste* typ_nr;
  // Modul w speichern
  startzeit[w-1] = st;
  st += laufzeit( w, re - li );
  breite[w-1] = re - li;
  linker[w-1] = li + 1;
  reset_kind( g, w );
  anz = 0; // Anzahl der Kinder
  while( 1 )
  {
    z = next_kind( g, w );
    if( z == 0 )
    {
      break;
    }
    if( indeg[z-1] == 1 )
    {
      anz++;
    }
    else
    {
      // Eingangsgrad später um 1 reduzieren
      if( add_anf_il( liste, z ) != 0 )
      {
        return -1;
      }
    }
  }
  if( anz == 0 )
  {
    return st;
  }
  reset_kind( g, w );
  typ_nr = swap_kind_get( g, w );
  return setze_zeiten_formel_bal2( g, w, indeg, typ, liste, n, laufzeit,
                                   startzeit, breite, linker, li, re,
                                   st, anz, &typ_nr );
}

// Optimierungsroutine; siehe oben
int optimiere_formel(
  double** startzeit,
  int**    breite,
  int**    linker,
  graph*   g,
  int      n,
  double   ((*laufzeit)( int modulnummer, int prozessorenanzahl )) )
{
  int i, m, akt, neu, z, li, re;
  double st;
  double* endzeit;
  int* indeg;
  int* typ;
  double* za;
  int anz;
  il_pos typ_nr;
  intliste liste[2];
  init_il( &(liste[0]) );
  init_il( &(liste[1]) );
  m = max_knoten( g );
  *startzeit = (double*)malloc( m * sizeof( double ) );
  if( *startzeit == NULL )
  {
    return 1;
  }
  *breite = (int*)malloc( m * sizeof( int ) );
  if( *breite == NULL )
  {
    return 1;
  }
  *linker = (int*)malloc( m * sizeof( int ) );
  if( *linker == NULL )
  {
    return 1;
  }
  endzeit = (double*)malloc( n * sizeof( double ) );
  if( endzeit == NULL )
  {
    return 1;
  }
  za = (double*)malloc( n * sizeof( double ) );
  if( za == NULL )
  {
    return 1;
  }
  indeg = (int*)malloc( m * sizeof( int ) );
  if( indeg == NULL )
  {
    return 1;
  }
  typ = (int*)malloc( m * n * sizeof( int ) );
  if( typ == NULL )
  {
    return 1;
  }
  anz = 0; // zählen der Listenelemente
  for( i=0; i<m; i++ )
  {
    indeg[i] = eing_grad( g, i+1 );
    if( indeg[i] == 0 )
    {
      if( add_anf_il( &(liste[0]), i+1 ) != 0 )
      {
        return 1;
      }
      anz++;
    }
  }
  st = 0.0; // bis dahin wurde eingepackt
  akt = 0;
  neu = 1;
  while( 1 )
  {
    reset_il( &(liste[akt]) );
    z = get_il( &(liste[akt]) );
    if( z == 0 ) // Liste ist leer
    {
      break;
    }
    get_pos_il( &(liste[akt]), &typ_nr );
    if( verein_modul_formel_bal2( g, &(liste[akt]), indeg, typ, n, za,
                                  laufzeit, anz, &typ_nr ) != 0 )
    {
      return 1;
    }
    li = 0;
    re = n;
    reset_il( &(liste[akt]) );
    st = setze_zeiten_formel_bal1( g, &(liste[akt]), indeg, typ,
                                   &(liste[neu]), n, laufzeit,
                                   *startzeit, *breite, *linker,
                                   li, re, st, anz );
    if( st < 0 )
    {
      return 1;
    }
    anz = 0;
    reset_il( &(liste[neu]) );
    // Eingangsgrade reduzieren und Wurzeln der neuen Bäume merken
    while( 1 )
    {
      z = get_il( &(liste[neu]) );
      if( z == 0 )
      {
        break;
      }
      if( indeg[z-1] != 1 )
      {
        indeg[z-1]--;
        del_il( &(liste[neu]) );
      }
      else
      {
        next_il( &(liste[neu]) );
        anz++;
      }
    }
    z = neu;
    neu = akt;
    akt = z;
  }
  free_il( &(liste[0]) );
  free_il( &(liste[1]) );
  free( za );
  free( indeg );
  free( endzeit );
  free( typ );
  return 0;
}

#endif


// ------------------ Dateiorientierte Routinen ------------------------
#ifndef FORMEL

int verein_modul_array(
  graph*  g,
  int     w,
  int*    indeg,
  int*    typ,
  int     n,
  double* za,
  double* t );

// berechnet die Zeiten za vom fiktiven Modul w (rekursiver Abstieg
// zum Ausbalancieren des Baumes)
// g ist der Abhängigkeitsgraph und indeg seine fiktiven Eingangsgrade;
// in typ wird die durchgeführte Vereinigung gespeichert: es sei z die
// Nummer des i-ten Kindes, so stehen ab typ[(z-1)*n] die
// Vereinigungsdaten des i-ten inneren Knoten bzgl. DFS des
// Konstruktionsbaumes der Kinder von w
// Fehler: 1->zu wenig Speicher
int verein_modul_array_bal1(
  graph*        g,
  int           w,
  int*          indeg,
  int*          typ,
  int           n,
  double*       za,
  double*       t,
  int           anz,
  knotenliste** typ_nr )
{
  double* zb;
  int z;
  if( anz == 1 )
  {
    do
    {
      z = next_kind( g, w );
    } while( indeg[z-1] != 1 );
    return verein_modul_array( g, z, indeg, typ, n, za, t );
  }
  zb = (double*)malloc( n * sizeof( double ) );
  if( zb == NULL )
  {
    return 1;
  }
  // suchen des Moduls für die Vereinigungsinformationen typ
  *typ_nr = swap_kind( g, w, *typ_nr );
  do
  {
    z = next_kind( g, w );
  } while( indeg[z-1] != 1 );
  *typ_nr = swap_kind( g, w, *typ_nr );
  if( verein_modul_array_bal1( g, w, indeg, typ, n, za, t,
                               anz - anz/2, typ_nr ) != 0 )
  {
    return 1;
  }
  if( verein_modul_array_bal1( g, w, indeg, typ, n, zb, t,
                               anz/2, typ_nr ) != 0 )
  {
    return 1;
  }
  verein_parallel( z, n, typ, za, zb );
  free( zb );
  return 0;
}

// berechnet die Zeiten za von den Modulen in der Liste und vereinigt
// sie parallel (rekursiver Abstieg zum Ausbalancieren des Baumes)
// g ist der Abhängigkeitsgraph und indeg seine fiktiven Eingangsgrade;
// in typ wird die durchgeführte Vereinigung gespeichert
// Fehler: 1->zu wenig Speicher
int verein_modul_array_bal2(
  graph*    g,
  intliste* liste,
  int*      indeg,
  int*      typ,
  int       n,
  double*   za,
  double*   t,
  int       anz,
  il_pos*   typ_nr )
{
  double* zb;
  int z;
  if( anz == 1 )
  {
    z = next_il( liste );
    return verein_modul_array( g, z, indeg, typ, n, za, t );
  }
  zb = (double*)malloc( n * sizeof( double ) );
  if( zb == NULL )
  {
    return 1;
  }
  // suchen des Moduls für die Vereinigungsinformationen typ
  swap_pos_il( liste, typ_nr );
  z = next_il( liste );
  swap_pos_il( liste, typ_nr );
  if( verein_modul_array_bal2( g, liste, indeg, typ, n, za, t,
                               anz - anz/2, typ_nr ) != 0 )
  {
    return 1;
  }
  if( verein_modul_array_bal2( g, liste, indeg, typ, n, zb, t,
                               anz/2, typ_nr ) != 0 )
  {
    return 1;
  }
  verein_parallel( z, n, typ, za, zb );
  free( zb );
  return 0;
}

// berechnet die Zeiten za vom fiktiven Modul w;
// g ist der Abhängigkeitsgraph und indeg seine fiktiven Eingangsgrade;
// in typ wird die durchgeführte Vereinigung gespeichert
// Fehler: 1->zu wenig Speicher
int verein_modul_array(
  graph*  g,
  int     w,
  int*    indeg,
  int*    typ,
  int     n,
  double* za,
  double* t )
{
  int i, z;
  int anz;
  knotenliste* typ_nr;
  reset_kind( g, w );
  anz = 0; // mögliche Kinder zählen
  while( 1 )
  {
    z = next_kind( g, w );
    if( z == 0 ) // keine Kinder mehr
    {
      break;
    }
    if( indeg[z-1] == 1 ) // verwendbares Kind
    {
      anz++;
    }
  }
  if( anz == 0 )
  { // Blatt in diesem Teilbaum erreicht; Modullaufzeiten zurückgeben
    for( i=0; i<n; i++ )
    {
      za[i] = t[ (w-1)*n + i ];
    }
    return 0;
  }
  reset_kind( g, w );
  typ_nr = swap_kind_get( g, w );
  if( verein_modul_array_bal1( g, w, indeg, typ, n, za, t,
                               anz, &typ_nr ) != 0 )
  {
    return 1;
  }
  // sequentielle Vereinigung mit Modul w
  for( i=0; i<n; i++ )
  {
    za[i] += t[ (w-1)*n + i ];
  }
  return 0;
}

double setze_zeiten_array(
  graph*    g,
  int       w,
  int*      indeg,
  int*      typ,
  intliste* liste,
  int       n,
  double*   t,
  double*   startzeit,
  int*      breite,
  int*      linker,
  int       li,
  int       re,
  double    st );

// arbeitet Elemente der nachfolgerliste baumartig ab und berechnet die
// Startzeiten
// Fehler: <0 zu wenig Speicher
double setze_zeiten_array_bal1(
  graph*    g,
  intliste* nachfolger,
  int*      indeg,
  int*      typ,
  intliste* liste,
  int       n,
  double*   t,
  double*   startzeit,
  int*      breite,
  int*      linker,
  int       li,
  int       re,
  double    st,
  int       anz )
{
  int z, h;
  double sr;
  if( anz == 1 ) // nur noch ein Kind
  {
    z = del_anf_il( nachfolger );
    return setze_zeiten_array( g, z, indeg, typ, liste, n, t, startzeit,
                               breite, linker, li, re, st );
  }
  z = next_il( nachfolger ); // Modul mit Vereinigungsinformationen
  if( typ[ (z-1)*n + re - li - 1 ] == 0 ) // seq. Vereinigung
  {
    st = setze_zeiten_array_bal1( g, nachfolger, indeg, typ, liste,
                                  n, t, startzeit, breite, linker,
                                  li, re, st, anz - anz/2 );
    if( st < 0 )
    {
      return -1;
    }
    return setze_zeiten_array_bal1( g, nachfolger, indeg, typ, liste,
                                    n, t, startzeit, breite, linker,
                                    li, re, st, anz/2 );
  }
  // parallele Vereinigung
  h = li + typ[ (z-1)*n + re - li - 1 ];
  sr = setze_zeiten_array_bal1( g, nachfolger, indeg, typ, liste, n, t,
                                startzeit, breite, linker, li, h, st,
                                anz - anz/2 );
  if( sr < 0 )
  {
    return -1;
  }
  st = setze_zeiten_array_bal1( g, nachfolger, indeg, typ, liste, n, t,
                                startzeit, breite, linker, h, re, st,
                                anz/2 );
  if( st < 0 )
  {
    return -1;
  }
  if( sr > st )
  {
    return sr;
  }
  return st;
}

// arbeitet die Kinder von baumartig ab und berechnet die Startzeiten
// Fehler: <0 zu wenig Speicher
double setze_zeiten_array_bal2(
  graph*        g,
  int           w,
  int*          indeg,
  int*          typ,
  intliste*     liste,
  int           n,
  double*       t,
  double*       startzeit,
  int*          breite,
  int*          linker,
  int           li,
  int           re,
  double        st,
  int           anz,
  knotenliste** typ_nr )
{
  int z, h;
  double sr;
  if( anz == 1 ) // nur noch ein Kind
  {
    do
    {
      z = next_kind( g, w );
    } while( indeg[z-1] != 1 );
    return setze_zeiten_array( g, z, indeg, typ, liste, n, t, startzeit,
                               breite, linker, li, re, st );
  }
  // Modul mit Vereinigungsinformationen suchen
  *typ_nr = swap_kind( g, w, *typ_nr );
  do
  {
    z = next_kind( g, w );
  } while( indeg[z-1] != 1 );
  *typ_nr = swap_kind( g, w, *typ_nr );
  if( typ[ (z-1)*n + re - li - 1 ] == 0 ) // seq. Vereinigung
  {
    st = setze_zeiten_array_bal2( g, w, indeg, typ, liste, n, t,
                                  startzeit, breite, linker, li,
                                  re, st, anz - anz/2, typ_nr );
    if( st < 0 )
    {
      return -1;
    }
    return setze_zeiten_array_bal2( g, w, indeg, typ, liste, n, t,
                                    startzeit, breite, linker, li, re,
                                    st, anz/2, typ_nr );
  }
  // parallele Vereinigung
  h = li + typ[ (z-1)*n + re - li - 1 ];
  sr = setze_zeiten_array_bal2( g, w, indeg, typ, liste, n, t,
                                startzeit, breite, linker, li, h, st,
                                anz - anz/2, typ_nr );
  if( sr < 0 )
  {
    return -1;
  }
  st = setze_zeiten_array_bal2( g, w, indeg, typ, liste, n, t,
                                startzeit, breite, linker, h, re, st,
                                anz/2, typ_nr );
  if( st < 0 )
  {
    return -1;
  }
  if( sr > st )
  {
    return sr;
  }
  return st;
}

// speichert das Ergebnis indem der rekursive Abstieg, wie in
// verein_modul berechnet, durchgeführt wird
// Rückgabe: >0 Endzeit des fiktiven Knotens w bei Startzeit st und
//              Breite re-li und linken Prozessor li
//           <0 zu wenig Speicher
double setze_zeiten_array(
  graph*    g,
  int       w,
  int*      indeg,
  int*      typ,
  intliste* liste,
  int       n,
  double*   t,
  double*   startzeit,
  int*      breite,
  int*      linker,
  int       li,
  int       re,
  double    st )
{
  int anz, z;
  knotenliste* typ_nr;
  // Modul w speichern
  startzeit[w-1] = st;
  st += t[ (w-1)*n + re - li - 1 ];
  breite[w-1] = re - li;
  linker[w-1] = li + 1;
  reset_kind( g, w );
  anz = 0; // Anzahl der Kinder
  while( 1 )
  {
    z = next_kind( g, w );
    if( z == 0 )
    {
      break;
    }
    if( indeg[z-1] == 1 )
    {
      anz++;
    }
    else
    {
      // Eingangsgrad später um 1 reduzieren
      if( add_anf_il( liste, z ) != 0 )
      {
        return -1;
      }
    }
  }
  if( anz == 0 )
  {
    return st;
  }
  reset_kind( g, w );
  typ_nr = swap_kind_get( g, w );
  return setze_zeiten_array_bal2( g, w, indeg, typ, liste, n, t,
                                  startzeit, breite, linker, li, re,
                                  st, anz, &typ_nr );
}

// Optimierungsroutine siehe oben
int optimiere_datei(
  char* ergebnis,
  char* daten,
  char* graph_datei )
{
  int i, j, m, n, akt, neu, z, li, re;
  double st, r;
  double* endzeit;
  double* t;
  double* startzeit;
  int* linker;
  int* breite;
  int* indeg;
  int* typ;
  int pos;
  double* za;
  int anz;
  il_pos typ_nr;
  intliste liste[2];
  FILE* fi;
  graph g;
  init_graph( &g );
  // Graph laden
  if( load_graph_rueck_mehr( graph_datei, &g, 2 ) != 0 )
  {
    return 3;
  }
  // Maximale Knotenanzahl ermitteln
  m = max_knoten( &g );
  // Berechnen der Lösung
  fi = fopen( daten, "r" );
  if( fi == NULL )
  {
    return 2;
  }
  // Bestimmen der Prozessorenanzahl
  pos = ftell( fi );
  n = 1;
  if( lese_next_double( &fi, &r ) != 0 )
  {
    return 2;
  }
  while( lese_next_double_in_line( &fi, &r ) == 0 )
  {
    n++;
  }
  fseek( fi, pos, SEEK_SET );
  // Hilfsvariablen
  t = (double*) malloc( sizeof( double ) * m * n );
  if( t == NULL )
  {
    return 1;
  }
  // Laufzeiten laden
  for( i=0; i<m; i++ )
  {
    if( lese_next_double( &fi, &(t[i*n]) ) != 0 )
    { // m ist nur eine obere Schranke
      break;
    }
    for( j=1; j<n; j++ )
    {
      if( lese_next_double_in_line( &fi, &(t[i*n+j]) ) != 0 )
      {
        return 2;
      }
    }
    lese_next_ret( &fi );
  }
  fclose( fi );
  init_il( &(liste[0]) );
  init_il( &(liste[1]) );
  startzeit = (double*)malloc( m * sizeof( double ) );
  if( startzeit == NULL )
  {
    return 1;
  }
  breite = (int*)malloc( m * sizeof( int ) );
  if( breite == NULL )
  {
    return 1;
  }
  linker = (int*)malloc( m * sizeof( int ) );
  if( linker == NULL )
  {
    return 1;
  }
  endzeit = (double*)malloc( n * sizeof( double ) );
  if( endzeit == NULL )
  {
    return 1;
  }
  za = (double*)malloc( n * sizeof( double ) );
  if( za == NULL )
  {
    return 1;
  }
  indeg = (int*)malloc( m * sizeof( int ) );
  if( indeg == NULL )
  {
    return 1;
  }
  typ = (int*)malloc( m * n * sizeof( int ) );
  if( typ == NULL )
  {
    return 1;
  }
  anz = 0; // zählen der Listenelemente
  for( i=0; i<m; i++ )
  {
    indeg[i] = eing_grad( &g, i+1 );
    if( indeg[i] == 0 )
    {
      if( add_anf_il( &(liste[0]), i+1 ) != 0 )
      {
        return 1;
      }
      anz++;
    }
  }
  st = 0.0; // bis dahin wurde eingepackt
  akt = 0;
  neu = 1;
  while( 1 )
  {
    reset_il( &(liste[akt]) );
    z = get_il( &(liste[akt]) );
    if( z == 0 ) // Liste ist leer
    {
      break;
    }
    get_pos_il( &(liste[akt]), &typ_nr );
    if( verein_modul_array_bal2( &g, &(liste[akt]), indeg, typ,
                                 n, za, t, anz, &typ_nr ) != 0 )
    {
      return 1;
    }
    li = 0;
    re = n;
    reset_il( &(liste[akt]) );
    st = setze_zeiten_array_bal1( &g, &(liste[akt]), indeg, typ,
                                  &(liste[neu]), n, t, startzeit,
                                  breite, linker, li, re, st, anz );
    if( st < 0 )
    {
      return 1;
    }
    anz = 0;
    reset_il( &(liste[neu]) );
    // Eingangsgrade reduzieren und Wurzeln der neuen Bäume merken
    while( 1 )
    {
      z = get_il( &(liste[neu]) );
      if( z == 0 )
      {
        break;
      }
      if( indeg[z-1] != 1 )
      {
        indeg[z-1]--;
        del_il( &(liste[neu]) );
      }
      else
      {
        next_il( &(liste[neu]) );
        anz++;
      }
    }
    z = neu;
    neu = akt;
    akt = z;
  }
  free_il( &(liste[0]) );
  free_il( &(liste[1]) );
  free( za );
  free( indeg );
  free( endzeit );
  free( typ );
  free_graph( &g );
  fi = fopen( ergebnis, "w" );
  if( fi == NULL )
  {
    return 4;
  }
  for( i=0; i<m; i++ )
  {
    if( save_text( &fi, "Modul " ) != 0 )
    {
      return 4;
    }
    if( save_int( &fi, i+1 ) != 0 )
    {
      return 4;
    }
    if( save_text( &fi, ": von " ) != 0 )
    {
      return 4;
    }
    if( save_double( &fi, startzeit[i] ) != 0 )
    {
      return 4;
    }
    if( save_text( &fi, " bis " ) != 0 )
    {
      return 4;
    }
    if( save_double( &fi, startzeit[i] + t[ i*n + breite[i]-1 ] ) != 0 )
    {
      return 4;
    }
    if( save_text( &fi, " auf Prozessoren " ) != 0 )
    {
      return 4;
    }
    if( save_int( &fi, linker[i] ) != 0 )
    {
      return 4;
    }
    if( save_text( &fi, " bis " ) != 0 )
    {
      return 4;
    }
    if( save_int_r( &fi, linker[i] + breite[i] - 1 ) != 0 )
    {
      return 4;
    }
  }
  fclose( fi );
  free( t );
  free( startzeit );
  free( breite );
  free( linker );
  return 0;
}

#endif
