// Optimierungsroutinen nonmoldable, beliebiger Modulabhängigkeitsgraph
// Verbreiterung der Module auf dem kritischen Pfad
//
// optimiert Module deren Laufzeiten in der Datei daten stehen
// (i-te Zeile j-te Spalte enthält Laufzeit vom i-ten Modul
//  mit j Prozessoren)
// oder mit Hilfe der übergebenen Funktion laufzeit
// (modulnummer>=1, prozessorenanzahl>=1) ermittelt werden können
//
// Aufruf:
//   optimiere_datei(
//     char* Ergebnisdatei,
//     char* Laufzeitendatei,
//     char* Graphendatei )
// oder:
//   optimiere_formel(
//     double** Startzeit der Module,
//     int**    Modulbreiten,
//     int**    linker Prozessor jedes Moduls,
//     graph*   Modulabhängigkeitsgraph,
//     int      Maximalprozessorenanzahl,
//     double   ((*laufzeit)( modulnummer, prozessorenanzahl )) )
//   Der Speicher für die Startzeiten, Breiten und linker Prozessor
//   wird reserviert
//
// Fehler: 1->zu wenig Speicher
//         2->Eingabedatei mit Laufzeiten nicht gefunden / fehlerhaft
//         3->Eingabedatei mit Graphen nicht gefunden / fehlerhaft /
//            Graph enthält Zyklus
//         4->kann Ergebnisdatei nicht anlegen / Schreibfehler

#include <float.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include "../prgx/graphen.h"
#include "../prgx/heap.h"
#include "../prgx/load.h"
#include "../prgx/save.h"
#include "../prgx/suchsort.h"

// ab dieser Differenz werden zwei Zeiten als gleich angesehen
#define EPSILON 0.000001


// berechnet die Zeit, die nach dem Start von Modul i noch mindestens
// benötigt wird
int berechne_restzeit_breit(
  graph*  g,
  double* t,
  int*    breite,
  double* rest,
  int     n,
  int     i )
{
  int j;
  reset_kind( g, i+1 );
  rest[i] = 0.0;
  j = next_kind( g, i+1 );
  while( j != 0 )
  {
    if( rest[j-1] < 0 )
    {
      berechne_restzeit_breit( g, t, breite, rest, n, j-1 );
    }
    if( rest[i] < rest[j-1] )
    {
      rest[i] = rest[j-1];
    }
    j = next_kind( g, i+1 );
  }
  // eigene Laufzeit wird immer noch benötigt
  rest[i] += t[ i*n + breite[i] ];
  return 0;
}

// berechnet die Zeit, die nach dem Start von Modul i noch mindestens
// benötigt wird
int berechne_restzeit_solo(
  graph*  g,
  double* t,
  double* rest,
  int     i )
{
  int j;
  reset_kind( g, i+1 );
  rest[i] = 0.0;
  j = next_kind( g, i+1 );
  while( j != 0 )
  {
    if( rest[j-1] < 0 )
    {
      berechne_restzeit_solo( g, t, rest, j-1 );
    }
    if( rest[i] < rest[j-1] )
    {
      rest[i] = rest[j-1];
    }
    j = next_kind( g, i+1 );
  }
  rest[i] += t[i]; // eigene Laufzeit wird immer noch benötigt
  return 0;
}


// ------------------ Dateiorientierte Routinen ------------------------
#ifndef FORMEL

// Optimierungsroutine siehe oben
int optimiere_datei(
  char* ergebnis,
  char* daten,
  char* graph_datei )
{
  graph g;
  int m, n, i, j, k, fl;
  int akt, pfad;
  double d, flae, best;
  double* t;
  double* rest;
  double* startzei[2];
  double* frueh;
  int* linke[2];
  int* breit[2];
  int* indeg;
  int* indeg_ret;
  double* p_ende;
  int pos;
  intheap bereit;
  intheap warte;
  FILE* fi;
  init_graph( &g );
  // Graph laden
  if( load_graph_rueck_mehr( graph_datei, &g, 2 ) != 0 )
  {
    return 3;
  }
  // Maximale Knotenanzahl ermitteln
  m = max_knoten( &g );
  fi = fopen( daten, "r" );
  if( fi == NULL )
  {
    return 2;
  }
  // Bestimmen der Prozessorenanzahl
  pos = ftell( fi );
  n = 1;
  if( lese_next_double( &fi, &d ) != 0 )
  {
    return 2;
  }
  while( lese_next_double_in_line( &fi, &d ) == 0 )
  {
    n++;
  }
  fseek( fi, pos, SEEK_SET );
  // Hilfsvariablen
  t = (double*) malloc( sizeof( double ) * m * n );
  if( t == NULL )
  {
    return 1;
  }
  breit[0] = (int*) malloc( sizeof( int ) * m );
  if( breit[0] == NULL )
  {
    return 1;
  }
  breit[1] = (int*) malloc( sizeof( int ) * m );
  if( breit[1] == NULL )
  {
    return 1;
  }
  linke[0] = (int*) malloc( sizeof( int ) * m );
  if( linke[0] == NULL )
  {
    return 1;
  }
  linke[1] = (int*) malloc( sizeof( int ) * m );
  if( linke[1] == NULL )
  {
    return 1;
  }
  rest = (double*) malloc( sizeof( double ) * m );
  if( rest == NULL )
  {
    return 1;
  }
  frueh = (double*) malloc( sizeof( double ) * m );
  if( frueh == NULL )
  {
    return 1;
  }
  startzei[0] = (double*) malloc( sizeof( double ) * m );
  if( startzei[0] == NULL )
  {
    return 1;
  }
  startzei[1] = (double*) malloc( sizeof( double ) * m );
  if( startzei[1] == NULL )
  {
    return 1;
  }
  indeg = (int*) malloc( sizeof( int ) * m );
  if( indeg == NULL )
  {
    return 1;
  }
  indeg_ret = (int*) malloc( sizeof( int ) * m );
  if( indeg_ret == NULL )
  {
    return 1;
  }
  p_ende = (double*) malloc( sizeof( double ) * n );
  if( p_ende == NULL )
  {
    return 1;
  }
  flae = 0.0;
  // Laufzeiten laden
  for( i=0; i<m; i++ )
  {
    breit[0][i] = -1; // minimale Breite suchen
    rest[i] = -1.0; // noch nicht berechnet
    frueh[i] = 0.0; // eher sollte man nichts starten
    d = DBL_MAX;
    for( j=0; j<n; j++ )
    {
      if( lese_next_double_in_line( &fi, &(t[i*n+j]) ) != 0 )
      {
        return 2;
      }
      if( t[i*n+j] != 0.0 )
      {
#ifndef MINFLAE
        if( breit[0][i] == -1 )
        {
          breit[0][i] = j; // minimale Breite festlegen
        }
#endif
        if( (j+1) * t[i*n+j] < d ) // minimale Fläche bestimmen
        {
          d = (j+1) * t[i*n+j];
#ifdef MINFLAE
          breit[0][i] = j;
#endif
        }
      }
    }
    flae += d;
    lese_next_ret( &fi );
  }
  fclose( fi );
  init_ih( &warte );
  init_ih( &bereit );
  // benötigte Restzeit bestimmen
  for( i=0; i<m; i++ )
  {
    if( rest[i] < 0 )
    {
      berechne_restzeit_breit( &g, t, breit[0], rest, n, i );
    }
    indeg[i] = eing_grad( &g, i+1 );
    indeg_ret[i] = indeg[i];
    if( indeg[i] == 0 )
    {
      if( add_max_indirect_ih( &bereit, i+1, rest ) != 0 )
      {
        return 1;
      }
    }
  }
  akt = 0;
  best = DBL_MAX;
  while( 1 )
  {
    // Startknoten des längsten Pfades bestimmen
    pfad = get_wurzel_ih( &bereit );
    // einpacken
    for( i=0; i<n; i++ )
    {
      p_ende[i] = 0.0;
    }
    j = 0;
    while( 1 )
    {
      i = del_max_wurzel_indirect_ih( &bereit, rest );
      if( i == 0 ) // Wurzel leer?
      {
        i = del_min_wurzel_indirect_ih( &warte, frueh ); // nachrücken
        if( i == 0 )
        {
          break; // nichts mehr da
        }
        if( add_max_indirect_ih( &bereit, i, rest ) != 0 )
        {
          return 1;
        }
        while( 1 )
        {
          k = get_wurzel_ih( &warte );
          if( k == 0 )
          {
            break;
          }
          // alle mit gleicher Startzeit übernehmen
          if( frueh[i-1] + EPSILON < frueh[k-1] )
          {
            break;
          }
          if( add_max_indirect_ih( &bereit, k, rest ) != 0 )
          {
            return 1;
          }
          del_min_wurzel_indirect_ih( &warte, frueh );
        }
        i = del_max_wurzel_indirect_ih( &bereit, rest );
      }
      if( breit[akt][i-1] == 0 )
      {
        if( p_ende[j] < frueh[i-1] ) // es entsteht ein Leerraum
        {
          for( k=0; k<j; k++ ) // ist das Einfügen weiter links möglich
          {
            if( p_ende[k] <= frueh[i-1] )
            {
              j = k;
              break;
            }
          }
          p_ende[j] = frueh[i-1];
        }
        startzei[akt][i-1] = p_ende[j];
        linke[akt][i-1] = j;
        p_ende[j] += t[ n*(i-1) ];
      }
      else
      {
        j = such_min_breit_limit( p_ende, n, breit[akt][i-1],
                                  frueh[i-1], &d );
        for( k=0; k<=breit[akt][i-1]; k++ )
        {
          p_ende[ j+k ] = d + t[ (i-1)*n + breit[akt][i-1] ];
        }
        startzei[akt][i-1] = d;
        linke[akt][i-1] = j;
      }
      reset_kind( &g, i );
      k = next_kind( &g, i );
      while( k != 0 )
      {
        (indeg[k-1])--;
        if( frueh[k-1] < p_ende[j] )
        {
          frueh[k-1] = p_ende[j];
        }
        if( indeg[k-1] == 0 )
        {
          if( add_min_indirect_ih( &warte, k, frueh ) != 0 )
          {
            return 1;
          }
        }
        k = next_kind( &g, i );
      }
      j = such_min( p_ende, n ); // niedrigster Prozessor
      // unter gleichen Bedingungen startbare Elemente hinzufügen
      while( 1 )
      {
        k = get_wurzel_ih( &warte );
        if( k == 0 )
        {
          break;
        }
        if( frueh[k-1] > p_ende[j] + EPSILON )
        {
          break;
        }
        if( add_max_indirect_ih( &bereit, k, rest ) != 0 )
        {
          return 0;
        }
        del_min_wurzel_indirect_ih( &warte, frueh );
      }
    }
    // Ergebnisgüte bestimmen
    j = such_max( p_ende, n );
    if( p_ende[j] * n <= flae * GUETEFAKTOR )
    {
      break;
    }
    if( p_ende[j] < best )
    {
      // Ergebnis retten
      best = p_ende[j];
      k = akt;
      akt = ( akt + 1 ) & 1;
      for( i=0; i<m; i++ )
      {
        breit[akt][i] = breit[k][i];
      }
    }
    // längsten Pfad suchen und die Breite (sofern möglich) aller
    // Knoten um Eins bzw. flächenminimal erhöhen
    fl = 0; // noch keine Erhöhung vorgenommen
    while( 1 )
    {
      if( breit[akt][ pfad-1 ] != n - 1 )
      {
        (breit[akt][ pfad-1 ])++;
#ifdef MINFLAE
        for( i=breit[akt][ pfad-1 ] + 1;  i<n;  i++ )
        {
          if( (i+1) * t[ n*(pfad-1)+i ] <
              ( ( breit[akt][ pfad-1 ] + 1 ) *
                t[ n*(pfad-1) + breit[akt][ pfad-1 ] ] ) )
          {
            breit[akt][ pfad-1 ] = i;
          }
        }
#endif
        fl = 1;
      }
      reset_kind( &g, pfad );
      i = next_kind( &g, pfad );
      if( i == 0 )
      {
        break; // Zielknoten
      }
      while( 1 )
      {
        j = next_kind( &g, pfad );
        if( j == 0 )
        {
          break;
        }
        if( rest[i-1] < rest[j-1] )
        {
          i = j;
        }
      }
      pfad = i;
    }
    if( fl == 0 )
    {
      akt = ( akt + 1 ) & 1; // wieder auf letztes Ergebnis schalten
      break; // keine Verbreiterung mehr möglich
    }
    for( i=0; i<m; i++ )
    {
      rest[i] = -1.0;
      frueh[i] = 0.0;
    }
    // benötigte Restzeit bestimmen
    for( i=0; i<m; i++ )
    {
      if( rest[i] < 0 )
      {
        berechne_restzeit_breit( &g, t, breit[akt], rest, n, i );
      }
      indeg[i] = indeg_ret[i];
      if( indeg[i] == 0 )
      {
        if( add_max_indirect_ih( &bereit, i+1, rest ) != 0 )
        {
          return 1;
        }
      }
    }
  }
  // Ergebnis speichern
  fi = fopen( ergebnis, "w" );
  if( fi == NULL )
  {
    return 4;
  }
  for( i=0; i<m; i++ )
  {
    if( save_text( &fi, "Modul " ) != 0 )
    {
      return 4;
    }
    if( save_int( &fi, i+1 ) != 0 )
    {
      return 4;
    }
    if( save_text( &fi, ": von " ) != 0 )
    {
      return 4;
    }
    if( save_double( &fi, startzei[akt][i] ) != 0 )
    {
      return 4;
    }
    if( save_text( &fi, " bis " ) != 0 )
    {
      return 4;
    }
    if( save_double(
          &fi, startzei[akt][i] + t[ i*n + breit[akt][i] ] ) != 0 )
    {
      return 4;
    }
    if( save_text( &fi, " auf Prozessoren " ) != 0 )
    {
      return 4;
    }
    if( save_int( &fi, linke[akt][i] + 1 ) != 0 )
    {
      return 4;
    }
    if( save_text( &fi, " bis " ) != 0 )
    {
      return 4;
    }
    if( save_int_r( &fi, linke[akt][i] + breit[akt][i] + 1 ) != 0 )
    {
      return 4;
    }
  }
  fclose( fi );
  free_ih( &bereit );
  free_ih( &warte );
  free( t );
  free( breit[0] );
  free( breit[1] );
  free( startzei[0] );
  free( startzei[1] );
  free( linke[0] );
  free( linke[1] );
  free( rest );
  free( frueh );
  free( indeg );
  free( indeg_ret );
  free( p_ende );
  free_graph( &g );
  return 0;
}

#endif


// ---------------------- Formelorientierte Routinen -------------------
#ifdef FORMEL

// Optimierungsroutine siehe oben
int optimiere_formel(
  double** startzeit,
  int**    breite,
  int**    linker,
  graph*   g,
  int      n,
  double   ((*laufzeit)( int modulnummer, int prozessorenanzahl )) )
{
  int m, i, j, k, fl;
  int akt, pfad;
  double d, dd, flae, best;
  double* t;
  double* rest;
  double* startzei[2];
  double* frueh;
  int* linke[2];
  int* breit[2];
  int* indeg;
  int* indeg_ret;
  double* p_ende;
  intheap bereit;
  intheap warte;
  // Maximale Knotenanzahl ermitteln
  m = max_knoten( g );
  // Hilfsvariablen
  t = (double*) malloc( sizeof( double ) * m * n );
  if( t == NULL )
  {
    return 1;
  }
  breit[0] = (int*) malloc( sizeof( int ) * m );
  if( breit[0] == NULL )
  {
    return 1;
  }
  breit[1] = (int*) malloc( sizeof( int ) * m );
  if( breit[1] == NULL )
  {
    return 1;
  }
  linke[0] = (int*) malloc( sizeof( int ) * m );
  if( linke[0] == NULL )
  {
    return 1;
  }
  linke[1] = (int*) malloc( sizeof( int ) * m );
  if( linke[1] == NULL )
  {
    return 1;
  }
  rest = (double*) malloc( sizeof( double ) * m );
  if( rest == NULL )
  {
    return 1;
  }
  frueh = (double*) malloc( sizeof( double ) * m );
  if( frueh == NULL )
  {
    return 1;
  }
  startzei[0] = (double*) malloc( sizeof( double ) * m );
  if( startzei[0] == NULL )
  {
    return 1;
  }
  startzei[1] = (double*) malloc( sizeof( double ) * m );
  if( startzei[1] == NULL )
  {
    return 1;
  }
  indeg = (int*) malloc( sizeof( int ) * m );
  if( indeg == NULL )
  {
    return 1;
  }
  indeg_ret = (int*) malloc( sizeof( int ) * m );
  if( indeg_ret == NULL )
  {
    return 1;
  }
  p_ende = (double*) malloc( sizeof( double ) * n );
  if( p_ende == NULL )
  {
    return 1;
  }
  flae = 0.0;
  // Laufzeiten mit minimaler Prozessoranzahl kopieren
  for( i=0; i<m; i++ )
  {
    rest[i] = -1.0; // noch nicht berechnet
    frueh[i] = 0.0; // eher sollte man nichts starten
    j = 0;
    do
    {
      j++;
      t[i] = (*laufzeit)( i+1, j );
    } while( t[i] == 0.0 );
    breit[0][i] = j; // am Anfang minimale Prozessorenanzahl verwenden
    d = t[i] * j;
    for( j++; j<=n; j++ )
    {
      dd = (*laufzeit)( i+1, j );
      if( d > dd * j )
      {
        d = dd * j;
#ifdef MINFLAE
        breit[0][i] = j;
        t[i] = dd;
#endif
      }
    }
    flae += d; // Mindestfläche
  }
  init_ih( &warte );
  init_ih( &bereit );
  // benötigte Restzeit bestimmen
  for( i=0; i<m; i++ )
  {
    if( rest[i] < 0 )
    {
      berechne_restzeit_solo( g, t, rest, i );
    }
    indeg[i] = eing_grad( g, i+1 );
    indeg_ret[i] = indeg[i];
    if( indeg[i] == 0 )
    {
      if( add_max_indirect_ih( &bereit, i+1, rest ) != 0 )
      {
        return 1;
      }
    }
  }
  akt = 0;
  best = DBL_MAX;
  while( 1 )
  {
    // Startknoten des längsten Pfades bestimmen
    pfad = get_wurzel_ih( &bereit );
    // einpacken
    for( i=0; i<n; i++ )
    {
      p_ende[i] = 0.0;
    }
    j = 0;
    while( 1 )
    {
      i = del_max_wurzel_indirect_ih( &bereit, rest );
      if( i == 0 ) // Wurzel leer?
      {
        i = del_min_wurzel_indirect_ih( &warte, frueh ); // nachrücken
        if( i == 0 )
        {
          break; // nichts mehr da
        }
        if( add_max_indirect_ih( &bereit, i, rest ) != 0 )
        {
          return 1;
        }
        while( 1 )
        {
          k = get_wurzel_ih( &warte );
          if( k == 0 )
          {
            break;
          }
          // alle mit gleicher Startzeit übernehmen
          if( frueh[i-1] + EPSILON < frueh[k-1] )
          {
            break;
          }
          if( add_max_indirect_ih( &bereit, k, rest ) != 0 )
          {
            return 1;
          }
          del_min_wurzel_indirect_ih( &warte, frueh );
        }
        i = del_max_wurzel_indirect_ih( &bereit, rest );
      }
      if( breit[akt][i-1] == 1 )
      {
        if( p_ende[j] < frueh[i-1] ) // es entsteht ein Leerraum
        {
          for( k=0; k<j; k++ ) // ist das Einfügen weiter links möglich
          {
            if( p_ende[k] <= frueh[i-1] )
            {
              j = k;
              break;
            }
          }
          p_ende[j] = frueh[i-1];
        }
        startzei[akt][i-1] = p_ende[j];
        linke[akt][i-1] = j + 1;
        p_ende[j] += t[i-1];
      }
      else
      {
        j = such_min_breit_limit( p_ende, n, breit[akt][i-1]-1,
                                  frueh[i-1], &d );
        for( k=0; k<breit[akt][i-1]; k++ )
        {
          p_ende[ j+k ] = d + t[i-1];
        }
        startzei[akt][i-1] = d;
        linke[akt][i-1] = j + 1;
      }
      reset_kind( g, i );
      k = next_kind( g, i );
      while( k != 0 )
      {
        (indeg[k-1])--;
        if( frueh[k-1] < p_ende[j] )
        {
          frueh[k-1] = p_ende[j];
        }
        if( indeg[k-1] == 0 )
        {
          if( add_min_indirect_ih( &warte, k, frueh ) != 0 )
          {
            return 1;
          }
        }
        k = next_kind( g, i );
      }
      j = such_min( p_ende, n ); // niedrigster Prozessor
      // unter gleichen Bedingungen startbare Elemente hinzufügen
      while( 1 )
      {
        k = get_wurzel_ih( &warte );
        if( k == 0 )
        {
          break;
        }
        if( frueh[k-1] > p_ende[j] + EPSILON )
        {
          break;
        }
        if( add_max_indirect_ih( &bereit, k, rest ) != 0 )
        {
          return 0;
        }
        del_min_wurzel_indirect_ih( &warte, frueh );
      }
    }
    // Ergebnisgüte bestimmen
    j = such_max( p_ende, n );
    if( p_ende[j] * n <= flae * GUETEFAKTOR )
    {
      break;
    }
    if( p_ende[j] < best )
    {
      // Ergebnis retten
      best = p_ende[j];
      k = akt;
      akt = ( akt + 1 ) & 1;
      for( i=0; i<m; i++ )
      {
        breit[akt][i] = breit[k][i];
      }
    }
    // längsten Pfad suchen und die Breite (sofern möglich) aller
    // Knoten um Eins bzw. fächenminimal erhöhen
    fl = 0; // noch keine Erhöhung vorgenommen
    while( 1 )
    {
      if( breit[akt][ pfad-1 ] != n )
      {
        (breit[akt][ pfad-1 ])++;
        t[ pfad-1 ] = (*laufzeit)( pfad, breit[akt][ pfad-1 ] );
#ifdef MINFLAE
        d = t[ pfad-1 ] * breit[akt][ pfad-1 ];
        for( i=breit[akt][ pfad-1 ] + 1;  i<=n;  i++ )
        {
          dd = (*laufzeit)( pfad, i );
          if( dd * i < d )
          {
            d = dd * i;
            t[ pfad-1 ] = dd;
            breit[akt][ pfad-1 ] = i;
          }
        }
#endif
        fl = 1;
      }
      reset_kind( g, pfad );
      i = next_kind( g, pfad );
      if( i == 0 )
      {
        break; // Zielknoten
      }
      while( 1 )
      {
        j = next_kind( g, pfad );
        if( j == 0 )
        {
          break;
        }
        if( rest[i-1] < rest[j-1] )
        {
          i = j;
        }
      }
      pfad = i;
    }
    if( fl == 0 )
    {
      akt = ( akt + 1 ) & 1; // wieder auf letztes Ergebnis schalten
      break; // keine Verbreiterung mehr möglich
    }
    for( i=0; i<m; i++ )
    {
      rest[i] = -1.0;
      frueh[i] = 0.0;
    }
    // benötigte Restzeit bestimmen
    for( i=0; i<m; i++ )
    {
      if( rest[i] < 0 )
      {
        berechne_restzeit_solo( g, t, rest, i );
      }
      indeg[i] = indeg_ret[i];
      if( indeg[i] == 0 )
      {
        if( add_max_indirect_ih( &bereit, i+1, rest ) != 0 )
        {
          return 1;
        }
      }
    }
  }
  // Ergebnis zurück geben
  k = ( akt + 1 ) & 1;
  free_ih( &bereit );
  free_ih( &warte );
  free( t );
  free( breit[k] );
  *breite = breit[akt];
  free( startzei[k] );
  *startzeit = startzei[akt];
  free( linke[k] );
  *linker = linke[akt];
  free( rest );
  free( frueh );
  free( indeg );
  free( indeg_ret );
  free( p_ende );
  return 0;
}

#endif
