#include <limits.h>
#include <malloc.h>
#include <stdlib.h>
#include "baum.h"

// Konstruktor - Ersatz; Parameter max gibt die Größe des Puffers für
// ungenutzte Knoten an (Reduzierung von malloc/free)
void init_size_baum( baum* b, int max )
{
  b->max = max;
  b->akt = 0;
  b->w = NULL;
}

// interne Hilfsroutine; löscht a (gepuffertes free)
void frei_help_baum( baum* b, ast* a )
{
  if( b->akt != b->max )
  {
    a->left = b->p;
    b->p = a;
    (b->akt)++;
    return;
  }
  free( a );
}

// interne Hilfsroutine; gibt neuen Ast
// Fehler: NULL->zu wenig Speicher
ast* neu_help_baum( baum* b )
{
  ast* a;
  if( b->akt == 0 )
  {
    return (ast*)malloc( sizeof( ast ) );
  }
  a = b->p;
  b->p = a->left;
  (b->akt)--;
  return a;
}

// rekursive Hilfsroutine zu del_?_baum (löscht a und alle seine
// Kinder, keine Änderung der Pointer zu a)
void del_all_baum_help( baum* b, ast* a )
{
  if( a->left != NULL )
  {
    del_all_baum_help( b, a->left );
  }
  if( a->right != NULL )
  {
    del_all_baum_help( b, a->right );
  }
  frei_help_baum( b, a );
}

// löscht a und alle Kinder von a
void del_all_baum( baum* b, ast* a )
{
  if( a == NULL )
  {
    return;
  }
  if( a->pred == NULL ) // a ist Wurzel
  {
    b->w = NULL;
  }
  else
  {
    if( a->pred->left == a ) // a hängt links
    {
      a->pred->left = NULL;
    }
    else
    {
      a->pred->right = NULL;
    }
  }
  del_all_baum_help( b, a );
}

// löscht Knoten a und den linken Teilbaum unterhalb von a;
// der rechte Unterbaum nimmt die Position von a ein
// Rückgabe: Knoten, der die Stelle von a eingenommen hat
//           (vorher a->right)
ast* del_l_baum( baum* b, ast* a )
{
  ast* c;
  // linken Unterbaum löschen
  if( a->left != NULL )
  {
    del_all_baum_help( b, a->left );
  }
  if( a->pred == NULL ) // a ist Wurzel
  {
    b->w = a->right;
    if( a->right != NULL )
    {
      a->right->pred = NULL;
    }
  }
  else
  {
    if( a->pred->left == a ) // a hängt links
    {
      a->pred->left = a->right;
      if( a->right != NULL )
      {
        a->right->pred = a->pred;
      }
    }
    else
    {
      a->pred->right = a->right;
      if( a->right != NULL )
      {
        a->right->pred = a->pred;
      }
    }
  }
  c = a->right;
  frei_help_baum( b, a );
  return c;
}

// Destruktor; Datenstruktur ist dann initialisiert,
// aber der Puffer ist leer
void free_baum( baum* b )
{
  ast* a;
  del_all_baum( b, b->w );
  while( b->akt != 0 )
  {
    a = b->p;
    b->p = a->left;
    free( a );
    (b->akt)--;
  }
}

// liefert die Wurzel des Baumes
ast* get_wurzel_baum( baum* b )
{
  return b->w;
}

// liefert einen neuen Knoten, der keine Nachfolger und Vorgänger hat
// Fehler: NULL->nicht genug Speicher
ast* new_baum( baum* b )
{
  ast* a;
  a = neu_help_baum( b );
  if( a == NULL )
  {
    return NULL;
  }
  a->left = NULL;
  a->right = NULL;
  a->pred = NULL;
  return a;
}

// fügt den Knoten c als linkes Kind von a ein;
// das linke Kind von a wird linkes Kind von c
// ist a == NULL so wird die Wurzel ersetzt
void insert_ast_ll_baum( baum* b, ast* a, ast* c )
{
  if( a != NULL )
  {
    c->left = a->left;
    a->left = c;
    c->pred = a;
    if( c->left != NULL )
    {
      c->left->pred = c;
    }
    return;
  }
  c->left = b->w;
  b->w = c;
  c->pred = NULL;
  if( c->left != NULL )
  {
    c->left->pred = c;
  }
}

// fügt einen Knoten als linkes Kind von a ein;
// das linke Kind von a wird linkes Kind vom neuen Knoten
// ist a == NULL so wird die Wurzel ersetzt
// Rückgabe: eingefügter Knoten
// Fehler: NULL->nicht genug Speicher; Baum ist unverändert
ast* insert_ll_baum( baum* b, ast* a )
{
  ast* c;
  c = new_baum( b );
  if( c == NULL )
  {
    return NULL;
  }
  insert_ast_ll_baum( b, a, c );
  return c;
}

// fügt den Knoten c als rechtes Kind von a ein; das rechte Kind von a
// wird rechtes Kind von c
// ist a == NULL so wird die Wurzel ersetzt
void insert_ast_rr_baum( baum* b, ast* a, ast* c )
{
  if( a != NULL )
  {
    c->right = a->right;
    a->right = c;
    c->pred = a;
    if( c->right != NULL )
    {
      c->right->pred = c;
    }
    return;
  }
  c->right = b->w;
  b->w = c;
  c->pred = NULL;
  if( c->right != NULL )
  {
    c->right->pred = c;
  }
}

// fügt einen Knoten als rechtes Kind von a ein; das rechte Kind von a
// wird rechtes Kind vom neuen Knoten
// ist a == NULL so wird die Wurzel ersetzt
// Rückgabe: eingefügter Knoten
// Fehler: NULL->nicht genug Speicher; Baum ist unverändert
ast* insert_rr_baum( baum* b, ast* a )
{
  ast* c;
  c = new_baum( b );
  if( c == NULL )
  {
    return NULL;
  }
  insert_ast_rr_baum( b, a, c );
  return c;
}

// fügt den Knoten c an die Stelle von a ein; das linke Kind von c
// wird a; a == NULL ist nicht zulässig
void insert_ast_l_baum( baum* b, ast* a, ast* c )
{
  if( a->pred == NULL ) // Wurzel ersetzen
  {
    c->left = a;
    a->pred = c;
    b->w = c;
    return;
  }
  if( a->pred->left == a ) // a ist linkes Kind
  {
    a->pred->left = c;
    c->pred = a->pred;
    c->left = a;
    a->pred = c;
    return;
  }
  // a ist rechtes Kind
  a->pred->right = c;
  c->pred = a->pred;
  c->left = a;
  a->pred = c;
}

// fügt einen Knoten an die Stelle von a ein; das linke Kind vom neuen
// Knoten wird a; a == NULL ist nicht zulässig
// Rückgabe: eingefügter Knoten
// Fehler: NULL->nicht genug Speicher; Baum ist unverändert
ast* insert_l_baum( baum* b, ast* a )
{
  ast* c;
  c = new_baum( b );
  if( c == NULL )
  {
    return NULL;
  }
  insert_ast_l_baum( b, a, c );
  return c;
}

// fügt den Knoten c an die Stelle von a ein; das rechte Kind von c
// wird a; a == NULL ist nicht zulässig
void insert_ast_r_baum( baum* b, ast* a, ast* c )
{
  if( a->pred == NULL ) // Wurzel ersetzen
  {
    c->right = a;
    a->pred = c;
    b->w = c;
    return;
  }
  if( a->pred->left == a ) // a ist linkes Kind
  {
    a->pred->left = c;
    c->pred = a->pred;
    c->right = a;
    a->pred = c;
    return;
  }
  // a ist rechtes Kind
  a->pred->right = c;
  c->pred = a->pred;
  c->right = a;
  a->pred = c;
}

// fügt einen Knoten an die Stelle von a ein; das rechte Kind vom neuen
// Knoten wird a; a == NULL ist nicht zulässig
// Rückgabe: eingefügter Knoten
// Fehler: NULL->nicht genug Speicher; Baum ist unverändert
ast* insert_r_baum( baum* b, ast* a )
{
  ast* c;
  c = new_baum( b );
  if( c == NULL )
  {
    return NULL;
  }
  insert_ast_r_baum( b, a, c );
  return c;
}

// hängt den Unterbaum mit Wurzel a rechts an den Knoten c
// !! c darf kein rechtes Kind haben !!
// !! c darf nicht unterhalb von a liegen !!
// !! c == NULL ist nicht zulässig !!
void move_r_baum( ast* a, ast* c )
{
  if( a->pred->left == a ) // linkes Kind
  {
    a->pred->left = NULL;
  }
  else
  {
    a->pred->right = NULL;
  }
  c->right = a;
  a->pred = c;
}

// hängt den Unterbaum mit Wurzel a links an den Knoten c
// !! c darf kein linkes Kind haben !!
// !! c darf nicht unterhalb von a liegen !!
// !! c == NULL ist nicht zulässig !!
void move_l_baum( ast* a, ast* c )
{
  if( a->pred->left == a ) // linkes Kind
  {
    a->pred->left = NULL;
  }
  else
  {
    a->pred->right = NULL;
  }
  c->left = a;
  a->pred = c;
}

// beschriftet den Knoten a mit dem Wert i
void set_baum( ast* a, int i )
{
  a->wert = i;
}

// liefert den Wert des Knotens a
int get_baum( ast* a )
{
  return a->wert;
}

// liefert das linke Kind von a
// Fehler: NULL->kein linkes Kind vorhanden
ast* left_baum( ast* a )
{
  return a->left;
}

// liefert das rechte Kind von a
// Fehler: NULL->kein rechtes Kind vorhanden
ast* right_baum( ast* a )
{
  return a->right;
}

// liefert den Vater von a
// Fehler: NULL->a ist Wurzel
ast* pred_baum( ast* a )
{
  return a->pred;
}
