#ifndef GRAPHEN_OP

#define GRAPHEN_OP

#ifdef DGRAPHEN_OP
#error "Entweder graphen.h oder dgraphen.h verwenden."
#endif

#include "listen.h"

typedef struct knotenliste
{
  struct knoten* knot;
  struct knotenliste* next;
} knotenliste;

typedef struct knoten
{
  int knotIdx; // Knotennummer
  knotenliste* forward; // Kanten zu Nachfolgerknoten
  knotenliste* backward; // Rückwärtskanten zu Vorgängerknoten
  knotenliste* currForward; // aktuell betrachteter Nachfolger
  knotenliste* currBackward; // aktuell betrachteter Vorgänger
} knoten;

typedef struct graph
{
  knoten** k; // Verweis auf einzelne Knoten (Feld)
  int max; // aktuelle Dimension der Feldes k
  int akt; // Anzahl der belegten Felder (größtes Feld)
  int* dfs; // Feld zum Speichern von Besucht-Informationen bei
  //           DFS-Läufen
  int dfsakt; // 0  dfs Feld nicht angelegt;
  //             >0 max. Größe des DFS-Feldes
  int dfsmax; // größter bisher genutzter Wert im DFS-Feld
  int dfsdiff; // die Werte dfsmax+1 bis dfsmax+dfsdiff werden genutzt
  knotenliste* kanpuf; // Puffer für Kanten
  int kanmax; // Maximalgröße des Kantenpuffers
  int kanakt; // aktueller Füllstand des Kantenpuffers
  knoten** knopuf; // Puffer für Knoten
  int knomax; // Maximalgröße des Knotenpuffers
  int knoakt; // aktueller Füllstand des Knotenpuffers
} graph;

// Konstruktor - Ersatz
void init_graph( graph* g );

// Destruktor - Ersatz; Graph ist dann wieder initialisiert;
// Puffer sind leer
void free_graph( graph* g );

// löscht den Graph (ohne Puffer zu leeren)
void del_graph( graph* g );

// lädt ein Graph g aus Datei fn und reserviert entsprechend Speicher
// (mehr steht als Empfehlung für die maximale
//  Knotenanzahl->Laufzeitminimierung)
// Dateiaufbau: der erste Eintrag in jeder Zeile stellt die aktuelle
// Knotennummer >= 1 dar; alle Knoten von denen eine Kante beim
// aktuellen Knoten eingehen folgen in der Zeile
// !! save_graph_rueck und anschließend load_graph_rueck ergibt nicht
//    die gleiche Reihenfolge in den Adjazenzlisten
// Fehler: 1->Datei nicht gefunden
//         2->zu wenig Speicher
int load_graph_rueck_mehr( char* fn, graph* g, int mehr );
int load_graph_rueck( char* fn, graph* g );

// speichert Graph; hinter jedem Knoten stehen seine Vorgänger
// !! save_graph_rueck und anschließend load_graph_rueck ergibt nicht
//    die gleiche Reihenfolge in den Adjazenzlisten
// Fehler: 1->kann Datei nicht anlegen / Schreibfehler
int save_graph_rueck( graph* g, char* fn );

// erzeugt einen Graphen der nur aus den Knoten 1 bis n besteht
// (wenn g hier vergrößert wird, so g->max = mehr wählen, um
//  nachfolgende Operationen optimal ausführen zu können)
// Fehler: 1->zu wenig Speicher
int create_knoten_graph_mehr( graph* g, int n, int mehr );
int create_knoten_graph( graph* g, int n );

// fügt Knoten ein (sofern er noch nicht vorhanden ist);
// mehr gibt bei einer Vergrößerung die neue gewünschte Obergrenze
// Fehler: 1->zu wenig Speicher
int add_knoten_mehr( graph* g, int i, int mehr );
int add_knoten( graph* g, int i );

// fügt eine Kante von i nach j ein; sollte Knoten i oder j nicht
// existieren, so wird er erstellt
// kein Test ob Kante schon existiert,
// mehr ist die Vergrößerungsempfehlung
// Fehler: 1->zu wenig Speicher
int add_kante_mehr( graph* g, int i, int j, int mehr );
int add_kante( graph* g, int i, int j );

// testet ob Knoten i existiert
// Rückgabe: 0->Knoten existiert
//           1->Knoten existiert nicht
int test_knoten( graph* g, int i );

// löscht Kante von i nach j
// Rückgabe: 0->Kante gelöscht
//           1->Anfangsknoten i nicht vorhanden
//           2->Endknoten j nicht vorhanden
//           3->keine Kante zwischen i und j
int del_kante( graph* g, int i, int j );

// löscht einen Knoten und alle seine Kanten
// Rückgabe: 0->Knoten gelöscht
//           1->Knoten nicht vorhanden
int del_knoten( graph* g, int i );

// liefert eine Obergrenze für die Knotenanzahl von g
int max_knoten( graph* g );

// testet ob i ein Zielknoten (Senke) ist
// Rückgabe: 0->i ist Senke
//           1->i ist keine Senke
//           2->i ist nicht vorhanden
int ziel_knoten( graph* g, int i );

// liefert den Eingangsgrad von Knoten i
// Rückgabe: -1->Knoten nicht vorhanden
int eing_grad( graph* g, int i );

// liefert den Ausgangsgrad von Knoten i
// Rückgabe: -1->Knoten nicht vorhanden
int ausg_grad( graph* g, int i );

// vor Verwendung von next_kind starten
// Rückgabe: 1->Knoten i nicht vorhanden
int reset_kind( graph* g, int i );

// stellt einen zuvor geretteten Zeiger k auf das nächste Kind des
// Knotens i wieder her
// Rückgabe: der aktuelle Zeiger auf das nächste Kind des Knotens i
knotenliste* swap_kind( graph* g, int i, knotenliste* k );

// liefert den aktuellen Zeiger auf das nächste Kind des Knotens i
knotenliste* swap_kind_get( graph* g, int i );

// gibt das nächste Kind des Knotens i; nach jeder Änderung des
// Graphen an diesem Knoten erst reset_kind starten!
// Rückgabe: 0 ->kein weiteres Kind vorhanden; Knoten i nicht vorhanden
//           >0->Nummer des Kindes
int next_kind( graph* g, int i );

// vor Verwendung von next_vorg starten
// Rückgabe: 1->Knoten i nicht vorhanden
int reset_vorg( graph* g, int i );

// gibt den nächsten Vorgänger des Knotens i ohne den Lesezeiger
// zu verändern (ein nachfolgender Aufruf von next_vorg liefert das
// gleiche Ergebnis)
// nach jeder Änderung des Graphen an diesem Knoten
// erst reset_vorg starten!
// Rückgabe: 0 ->kein weiterer Vorgänger vorhanden;
//               Knoten i nicht vorhanden
//           >0->Nummer des Vorgängers
int next_vorg( graph* g, int i );

// erzeugt einen zufälligen Graphen mit knoten Knoten und kanten Kanten
// es gibt nur Kanten von i nach i+1, ..., i+seq
// sollte Kantenanzahl größer als die Anzahl der möglichen Kanten
// sein, wird sie reduziert
// Fehler: 1->zu wenig Speicher
int rnd_graph( graph* g, int kanten, int seq, int knoten, int mehr );

// bestimmt die Tiefe der einzelnen Knoten;
// Speicher für tief_num (max_knoten+1 Einträge) muß reserviert sein
// in tief_num[0] steht die größte vergebene tief_nummer;
// in tief_num[i] die von Knoten i
// Fehler: 1->Graph enthält Zyklus
int tiefe( graph* g, int* tief_num );

// bestimmt die Höhe der einzelnen Knoten, also die Tiefen vom Graph
// mit umgedrehten Kanten; hoch_num[Zielknoten] = 1
// Speicher für hoch_num (max_knoten+1 Einträge) muß reserviert sein
// in hoch_num[0] steht die größte vergebene hoch_nummer;
// in hoch_num[i] die von Knoten i
// Fehler: 1->Graph enthält Zyklus
int hoehe( graph* g, int* hoch_num );

#endif
