#include <float.h>
#include <limits.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include "convert.h"
#include "save.h"
#include "zufall.h"


// erzeugt gleichverteilte ganze Zahl im Intervall [a, b]
int gleich_int( int a, int b )
{
  long c, d, e, f, g;
  g = b - a;
  do
  {
    c = 0;
    d = 1;
    e = g;
    while( 1 )
    {
      if( e <= RAND_MAX )
      {
        do
        {
          f = rand();
        } while( f >=
                 ( e + 1 ) *
                 ( ( (long)RAND_MAX + 1 ) / ( e + 1 ) ) );
        c += d * ( f % ( e + 1 ) );
        break;
      }
      c += d * rand();
      d *= (long)RAND_MAX + 1;
      e /= (long)RAND_MAX + 1;
    }
  } while( c > g );
  return (int)c + a;
}

// erzeugt gleichverteilte gebrochene Zahl im Intervall [a, b]
double gleich_double( double a, double b )
{
  double c;
  c = (double)gleich_int( 0, INT_MAX );
  c *= b - a;
  c /= (double)INT_MAX;
  c += a;
  return c;
}

// erzeugt exponentialverteilte gebrochene Zahl im Intervall [a, b]
// mit Parameter lambda (ist lambda < 0, so entsteht eine
// gespiegelte Kurve)
double exponential_double( double a, double b, double lambda )
{
  double abslambda, expa, c;
  abslambda = ( lambda > 0 ) ? lambda : -lambda;
  expa = exp( -abslambda * a );
  c = -1.0 / abslambda * log( expa + gleich_double( 0.0, 1.0 ) *
                              ( exp( -b * abslambda ) - expa ) );
  return ( lambda > 0 ) ? c : a + b - c;
}

// erzeugt exponentialverteilte ganze Zahl im Intervall [a, b] mit
// Parameter lambda (ist lambda < 0, so entsteht eine gespiegelte Kurve)
int exponential_int(
  int    a,
  int    b,
  double lambda )
{
  int c;
  do
  {
    c = (int)exponential_double( (double)a, (double)(b + 1),
                                 lambda );
  } while( c > b );
  return c;
}

// erzeugt normalverteilte gebrochene Zahl im Intervall [a, b] mit
// Parametern alpha und sigma
double normal_double(
  double a,
  double b,
  double alpha,
  double sigma )
{
#define STUETZ 100
#define PI 3.1415926535
  double hoehe[ STUETZ + 1 ]; // Funktionswerte der Dichtefunktion
  double c, d, e, f, g;
  int i;
  c = sqrt( 2 * PI ) * sigma;
  d = -2.0 * sigma * sigma;
  e = a - alpha; // aktuelle Position immer um alpha verringert
  f = (b - a) / ((double)(STUETZ)); // Abstand der Stützpfeiler
  g = 0.0; // Fläche im Intervall [a, b) unter Dichtefunktion
  for( i=0; i<=STUETZ; i++ )
  {
    hoehe[i] = exp( e * e / d ) / c;
    g += hoehe[i]; // gleich Trapezsummen bilden
    e += f;
  }
  g -= 0.5 * ( hoehe[0] + hoehe[ STUETZ ] ); // das wurde zuviel addiert
  g *= f;
  g *= gleich_double( 0.0, 1.0 ); // gesuchte Fläche
  c = 0.0; // Fläche vor dem (i-1)-ten Trapez
  d = 0.5 * ( hoehe[0] + hoehe[1] ) * f; // Fläche vor dem i-ten Trapez
  i = 1;
  while( d < g )
  {
    c = d;
    d += 0.5 * ( hoehe[i] + hoehe[ i + 1 ] ) * f;
    i++;
  }
  if( hoehe[ i - 1 ] == hoehe[i] )
  {
    return ( g - c ) / hoehe[ i - 1 ] + a + f * ( i - 1 );
  }
  d = f / ( hoehe[i] - hoehe[ i - 1 ] );
  if( hoehe[ i - 1 ] < hoehe[i] )
  {
    return a + f * ( i - 1 ) - d * hoehe[ i - 1 ] +
      sqrt( ( d * hoehe[ i - 1 ] ) * ( d * hoehe[ i - 1 ] ) +
            2 * d * ( g - c ) );
  }
  return a + f * ( i - 1 ) - d * hoehe[ i - 1 ] -
    sqrt( ( d * hoehe[ i - 1 ] ) * ( d * hoehe[ i - 1 ] ) +
          2 * d * ( g - c ) );
}

// erzeugt normalverteilte ganze Zahl im Intervall [a, b] mit
// Parametern alpha und sigma
int normal_int(
  int    a,
  int    b,
  double alpha,
  double sigma )
{
  int c;
  do
  {
    c = (int)normal_double( (double)a, (double)(b + 1),
                            alpha, sigma );
  } while( c > b );
  return c;
}

// liefert entsprechend der Verteilung eine ganze Zufallszahl
int zufall_int( verteilung_int* v )
{
  switch( v->typ )
  {
    case gleich :
      return gleich_int( v->a, v->b );
    case exponential :
      return exponential_int( v->a, v->b, v->p1 );
    case normal :
      return normal_int( v->a, v->b, v->p1, v->p2 );
    default : // konstant
      return v->a;
  }
}

// liefert entsprechend der Verteilung eine ganze Zufallszahl
// die in der double-verteilung angegebenen Werte sollten ca.
// ganze Zahlen (0.5 bis 1.4999999 wird als 1 gelesen) sein
int zufall_int_from_double( verteilung_double* v )
{
  switch( v->typ )
  {
    case gleich :
      return gleich_int( ((int)(v->a + 0.5)),
                         ((int)(v->b + 0.5)) );
    case exponential :
      return exponential_int( ((int)(v->a + 0.5)),
                              ((int)(v->b + 0.5)), v->p1 );
    case normal :
      return normal_int( ((int)(v->a + 0.5)),
                         ((int)(v->b + 0.5)),
                         v->p1, v->p2 );
    default : // konstant
      return ((int)(v->a + 0.5));
  }
}

// liefert entsprechend der Verteilung eine ganze Zufallszahl
// Variablen werden nur am 1. Buchstaben erkannt
// (Unterscheidung von groß/klein)
// in var_wert[i] steht die Belegung von var_name[i]
// Fehler: 1->Syntax im String falsch
//         2->verwendete Variable unbekannt
int zufall_int_var(
  verteilung_text* v,
  char*            var_name,
  int*             var_wert,
  int*             erg )
{
  int i;
  int a, b;
  double p1, p2;
  i = convert_string2int_var( v->a, var_name, var_wert, &a );
  if( i != 0 )
  {
    return i;
  }
  switch( v->typ )
  {
    case gleich :
      i = convert_string2int_var( v->b, var_name, var_wert, &b );
      if( i != 0 )
      {
        return i;
      }
      *erg = gleich_int( a, b );
      return 0;
    case exponential :
      i = convert_string2int_var( v->b, var_name, var_wert, &b );
      if( i != 0 )
      {
        return i;
      }
      i = convert_string2double_varint( v->p1, var_name,
                                        var_wert, &p1 );
      if( i != 0 )
      {
        return i;
      }
      *erg = exponential_int( a, b, p1 );
      return 0;
    case normal :
      i = convert_string2int_var( v->b, var_name, var_wert, &b );
      if( i != 0 )
      {
        return i;
      }
      i = convert_string2double_varint( v->p1, var_name,
                                        var_wert, &p1 );
      if( i != 0 )
      {
        return i;
      }
      i = convert_string2double_varint( v->p2, var_name,
                                        var_wert, &p2 );
      if( i != 0 )
      {
        return i;
      }
      *erg = normal_int( a, b, p1, p2 );
      return 0;
    default : // konstant
      *erg = a;
      return 0;
  }
}

// liefert entsprechend der Verteilung eine gebrochene Zufallszahl
double zufall_double( verteilung_double* v )
{
  switch( v->typ )
  {
    case gleich :
      return gleich_double( v->a, v->b );
    case exponential :
      return exponential_double( v->a, v->b, v->p1 );
    case normal :
      return normal_double( v->a, v->b, v->p1, v->p2 );
    default : // konstant
      return v->a;
  }
}

// gibt die größtmögliche Zufallszahl einer Verteilung zurück
int zufall_max_int( verteilung_int* v )
{
  if( v->typ == konstant )
  {
    return v->a;
  }
  return v->b;
}

// Konstruktor für Variablen vom Typ verteilung_text
void init_verteilung_text( verteilung_text* v )
{
  v->a = NULL;
  v->b = NULL;
  v->p1 = NULL;
  v->p2 = NULL;
}

// Destruktor für Variablen vom Typ verteilung_text;
// v ist dann wieder initialisiert
void free_verteilung_text( verteilung_text* v )
{
  if( v->a != NULL )
  {
    free( v->a );
    v->a = NULL;
  }
  if( v->b != NULL )
  {
    free( v->b );
    v->b = NULL;
  }
  if( v->p1 != NULL )
  {
    free( v->p1 );
    v->p1 = NULL;
  }
  if( v->p2 != NULL )
  {
    free( v->p2 );
    v->p2 = NULL;
  }
}
